BROKER_URL = "redis://127.0.0.1:6379/3"

CELERYD_CONCURRENCY = 1
CELERYD_PREFETCH_MULTIPLIER = 1

CELERY_SEND_EVENTS = 1
CELERY_RESULT_BACKEND = BROKER_URL
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TASK_SERIALIZER = 'json'
CELERY_ACCEPT_CONTENT = ['json', 'msgpack', 'yaml']
CELERY_IMPORTS = ('cw_celerytask_helpers.redislogger', 'demotasks')

CUBICWEB_CELERYTASK_LOG_ALL = True
CUBICWEB_CELERYTASK_REDIS_URL = BROKER_URL



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import traceback

import celery
from celery.events import EventReceiver
from kombu import Connection as BrokerConnection

from cubicweb.toolsutils import Command
from cubicweb.server.serverconfig import ServerConfiguration
from cubicweb.cwctl import CWCTL


class CeleryMonitorCommand(Command):
    """Synchronize celery task statuses"""

    name = 'celery-monitor'
    arguments = '<instance>'
    min_args = max_args = 1
    options = (
        ('loglevel',
         {'short': 'l', 'type': 'choice', 'metavar': '<log level>',
          'default': 'info', 'choices': ('debug', 'info', 'warning', 'error')},
         ),
    )

    def run(self, args):
        from cubicweb import repoapi
        from cubicweb.cwctl import init_cmdline_log_threshold
        config = ServerConfiguration.config_for(args[0])
        config.global_set_option('log-file', None)
        config.log_format = '%(levelname)s %(name)s: %(message)s'
        init_cmdline_log_threshold(config, self['loglevel'])
        repo = repoapi.get_repository(config=config)
        repo.hm.call_hooks('server_maintenance', repo=repo)
        self.repo = repo
        with repo.internal_cnx() as cnx:
            self.on_monitor_start(cnx)
        self.celery_monitor()

    @staticmethod
    def on_monitor_start(cnx):
        for adapter in cnx.vreg['adapters']['ICeleryTask']:
            adapter.on_monitor_start(cnx)
        cnx.commit()

    def on_event(self, event):
        with self.repo.internal_cnx() as cnx:
            try:
                for adapter in cnx.vreg['adapters']['ICeleryTask']:
                    adapter.on_event(cnx, event)
                cnx.commit()
            except Exception as exc:
                self.error('<CeleryMonitorCommand> '
                           'Unexpected error on event %s:\n%s',
                           event, exc)
                cnx.roolback()

    def celery_monitor(self):
        result_backend = celery.current_app.conf['CELERY_RESULT_BACKEND']
        while True:
            try:
                with BrokerConnection(result_backend) as conn:
                    recv = EventReceiver(
                        conn,
                        handlers={
                            'task-failed': self.on_event,
                            'task-succeeded': self.on_event,
                            'task-sent': self.on_event,
                            'task-received': self.on_event,
                            'task-revoked': self.on_event,
                            'task-started': self.on_event,
                            # '*': self.on_event,
                        })
                    recv.capture(limit=None, timeout=None)
            except (KeyboardInterrupt, SystemExit):
                traceback.print_exc()
                break


CWCTL.register(CeleryMonitorCommand)



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-celerytask entity's classes"""
import six
import warnings

import celery
from celery.result import AsyncResult, from_serializable

from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.view import EntityAdapter
from cubicweb.predicates import is_instance
from cubicweb.server.hook import DataOperationMixIn, Operation

from cw_celerytask_helpers import redislogger as loghelper

from cubes.celerytask import STATES, FINAL_STATES

_ = six.text_type

_TEST_TASKS = {}


def get_tasks():
    """Return tasks to be run (for use in cubicweb test mode)"""
    return _TEST_TASKS


def run_all_tasks(cnx=None):
    """Run all pending tasks (for use in cubicweb test mode)"""
    if cnx is None:
        warnings.warn('cnx argument should be specified to enable task '
                      'workflow synchronisation', DeprecationWarning,
                      stacklevel=2)
    results = {}
    for task_eid, task in _TEST_TASKS.items():
        results[task_eid] = task.delay()

    if cnx is not None:
        if celery.current_app.conf.CELERY_ALWAYS_EAGER:
            for task_eid, result in results.items():
                wf = cnx.entity_from_eid(task_eid).cw_adapt_to('IWorkflowable')
                transition = {
                    STATES.SUCCESS: 'finish',
                    STATES.FAILURE: 'fail',
                }[result.state]
                wf.fire_transition(transition, result.traceback)
        else:
            from cubes.celerytask.ccplugin import CeleryMonitorCommand
            CeleryMonitorCommand.on_monitor_start(cnx)
    return results


def start_async_task(cnx, task, *args, **kwargs):
    """Create and start a new task

    `task` can be either a task name, a task object or a task signature
    """
    task_name = six.text_type(celery.signature(task).task)
    entity = cnx.create_entity('CeleryTask', task_name=task_name)
    entity.cw_adapt_to('ICeleryTask').start(task, *args, **kwargs)
    return entity


def task_in_backend(task_id):
    app = celery.current_app
    if app.conf.CELERY_ALWAYS_EAGER:
        return False
    else:
        backend = app.backend
        return backend.get(backend.get_key_for_task(task_id)) is not None


class StartCeleryTaskOp(DataOperationMixIn, Operation):

    def postcommit_event(self):
        global _TEST_TASKS
        if self.cnx.vreg.config.mode == 'test':
            # In test mode, task should run explicitly with run_all_tasks()
            _TEST_TASKS = self.cnx.transaction_data.get('celerytask', {})
        else:
            for eid in self.get_data():
                task = self.cnx.transaction_data.get('celerytask', {}).get(eid)
                if task is not None:
                    task.delay()


class CeleryTask(AnyEntity):
    __regid__ = 'CeleryTask'
    fetch_attrs, cw_fetch_order = fetch_config(('task_name',))

    def dc_title(self):
        return self.task_name

    def dc_long_title(self):
        adapted = self.cw_adapt_to('ICeleryTask')
        state, finished = adapted.state, adapted.finished
        title = self.task_name or self._cw._('subtask')
        if finished:
            title = '%s (%s)' % (title, self._cw._(state))
        return title

    @property
    def progress(self):
        yield self.cw_adapt_to('ICeleryTask').progress
        for subtask in self.reverse_parent_task:
            yield subtask.progress

    @property
    def parent_tasks(self):
        yield self
        for task in self.parent_task:
            for ptask in task.parent_tasks:
                yield ptask


class ICeleryTask(EntityAdapter):
    __regid__ = 'ICeleryTask'
    __abstract__ = True

    def start(self, name, *args, **kwargs):
        eid = self.entity.eid
        task = self.get_task(name, *args, **kwargs)
        self._cw.transaction_data.setdefault('celerytask', {})[eid] = task
        StartCeleryTaskOp.get_instance(self._cw).add_data(eid)

    def get_task(self, name, *args, **kwargs):
        """Should return a celery task / signature or None

        This method run in a precommit event
        """
        return celery.signature(name, args=args, kwargs=kwargs)

    @staticmethod
    def on_event(cnx, event):
        """Triggered by celery-monitor"""
        pass

    @staticmethod
    def on_monitor_start(cnx):
        """Triggered by celery-monitor"""
        pass

    @property
    def task_id(self):
        raise NotImplementedError

    @property
    def task_name(self):
        raise NotImplementedError

    @property
    def logs(self):
        return loghelper.get_task_logs(self.task_id) or b''

    @property
    def result(self):
        return AsyncResult(self.task_id)

    @property
    def progress(self):
        if celery.current_app.conf.CELERY_ALWAYS_EAGER:
            return 1.
        result = self.result
        if result.info and 'progress' in result.info:
            return result.info['progress']
        elif self.entity.reverse_parent_task:
            children = self.entity.reverse_parent_task
            return sum(child.cw_adapt_to('ICeleryTask').progress
                       for child in children) / len(children)
        elif result.state == STATES.SUCCESS:
            return 1.
        else:
            return 0.

    @property
    def state(self):
        return self.result.state

    @property
    def finished(self):
        return self.state in FINAL_STATES


def get_task_id(task):
    if hasattr(task, 'freeze'):
        # false for ASyncResult
        result = task.freeze()
    else:
        result = task
    if hasattr(result, 'task_id'):
        return result.task_id
    elif hasattr(result, 'id'):
        return result.id
    else:
        # Group
        return get_task_id(task.tasks[-1])


class CeleryTaskAdapter(ICeleryTask):
    """Base adapter that store task call args in the transaction"""

    __select__ = ICeleryTask.__select__ & is_instance('CeleryTask')

    tr_map = {
        'task-failed': 'fail',
        'task-succeeded': 'finish',
        'task-received': 'enqueue',
        'task-started': 'start',
        'task-revoked': 'fail',
    }

    def attach_task(self, task, seen, parent=None):
        task_id = six.text_type(get_task_id(task))
        if parent is None:
            parent = self.entity
        if self.entity.task_id is None:
            self.entity.cw_set(task_id=task_id)
        elif task_id not in seen:
            task_name = six.text_type(task.task)
            parent = self._cw.create_entity('CeleryTask',
                                            task_id=six.text_type(task_id),
                                            task_name=task_name,
                                            parent_task=parent)
        seen.add(task_id)
        if hasattr(task, 'body'):
            self.attach_task(task.body, seen, parent)
        if hasattr(task, 'tasks'):
            for subtask in task.tasks:
                self.attach_task(subtask, seen, parent)

    def get_task(self, name, *args, **kwargs):
        task = super(CeleryTaskAdapter, self).get_task(
            name, *args, **kwargs)
        self.attach_task(task, set())
        return task

    @property
    def task_id(self):
        return self.entity.task_id

    @property
    def task_name(self):
        return self.entity.task_name

    def attach_result(self, result):
        def tree(result, seen=None):
            if seen is None:
                seen = set()
            if result.parent:
                for r in tree(result.parent, seen):
                    yield r
            for child in result.children or []:
                for r in tree(child, seen):
                    yield r

            if isinstance(result, AsyncResult):
                try:
                    rresult = from_serializable(result.result)
                except:
                    CeleryTaskAdapter.info(
                        'Cannot deserialize task result %s: %s',
                        result, result.result)
                else:
                    for r in tree(rresult, seen):
                        yield r

                if result.task_id not in seen:
                    seen.add(result.task_id)
                yield result

        for asr in tree(result):
            task_id = six.text_type(asr.id)
            rset = self._cw.find('CeleryTask', task_id=task_id)
            if rset:
                cwtask = rset.one()
            else:
                self.info("create a CeleryTask for %s" % task_id)
                cwtask = self._cw.create_entity(
                    'CeleryTask',
                    task_name=six.text_type(''),
                    task_id=six.text_type(task_id))
            if not cwtask.parent_task and self.entity is not cwtask:
                self.info('Set %s parent_task to %s (%s)', cwtask.task_id,
                          self.entity, self.entity.task_id)
                cwtask.cw_set(parent_task=self.entity)

    @staticmethod
    def on_event(cnx, event):
        # handle weaving of parent_task relations
        if event['type'] in ('task-received', 'task-sent'):
            # may create the cw task entity, if needed
            rset = cnx.find('CeleryTask', task_id=event['uuid'])
            if not rset:  # and 'name' in event:
                CeleryTaskAdapter.info("create CeleryTask for %s (%s)",
                                       event['uuid'], event['name'])
                entity = cnx.create_entity(
                    'CeleryTask',
                    task_id=six.text_type(event['uuid']),
                    task_name=six.text_type(event['name']),
                    )
                cnx.commit()
            else:
                entity = rset.one()
                CeleryTaskAdapter.info(
                    "using existing CeleryTask %s for %s (%s)",
                    entity, event['uuid'], event['name'])
            if entity.task_name != event['name']:
                if entity.task_name:
                    CeleryTaskAdapter.warning(
                        '<CeleryTask %s (task_id %s)> already has a name: %s',
                        entity.eid, entity.task_id, entity.task_name)
                CeleryTaskAdapter.info('set %s name to %s', entity.task_id,
                                       event['name'])
                entity.cw_set(task_name=event['name'])
            cnx.commit()

        elif event['type'] == 'task-succeeded':
            rset = cnx.find('CeleryTask', task_id=event['uuid'])
            if rset:
                asresult = AsyncResult(event['uuid'])
                CeleryTaskAdapter.info(
                    "Task %s:\n  asresult=%s\n  event.result=%s",
                    event['uuid'], asresult.result, event['result'])
                root = rset.one()
                root.cw_adapt_to('ICeleryTask').attach_result(asresult)
                cnx.commit()
            else:
                CeleryTaskAdapter.warning(
                    'Cannot find a cw entity for %s', event['uuid'])

        # manage workflow
        tr_map = CeleryTaskAdapter.tr_map
        if event['type'] in tr_map:
            rset = cnx.find('CeleryTask', task_id=event['uuid'])
            if not rset:
                CeleryTaskAdapter.warning(
                    '<CeleryTask (task_id %s)> not found in database; '
                    'cannot set it\'s state',
                    event['uuid'])
                return
            entity = rset.one()
            transition = tr_map[event['type']]
            CeleryTaskAdapter.fire_task_transition(
                entity, transition, event.get('exception'))

    @staticmethod
    def on_monitor_start(cnx):
        for task_eid, task_id in cnx.execute((
            'Any T, TID WHERE '
            'T is CeleryTask, T task_id TID, T in_state S, '
            'S name in ("waiting", "queued", "running")'
        )):
            result = AsyncResult(task_id)
            transition = {
                STATES.SUCCESS: 'finish',
                STATES.FAILURE: 'fail',
                STATES.STARTED: 'running',
            }.get(result.state)
            if transition is not None:
                CeleryTaskAdapter.fire_task_transition(
                    cnx.entity_from_eid(task_eid),
                    transition, result.traceback)

    @staticmethod
    def fire_task_transition(cnx, task, transition, traceback):
        try:
            wf = task.cw_adapt_to('IWorkflowable')
            wf.fire_transition_if_possible(
                transition, traceback)
            cnx.commit()
            CeleryTaskAdapter.info('<CeleryTask %s (task_id %s)> %s',
                                   task.eid, task.task_id, transition)
        except Exception as exc:
            CeleryTaskAdapter.error(
                '<CeleryTask %s (task_id %s)> failed to fire %s:\n%s',
                task.eid, task.task_id, transition, exc)
            cnx.rollback()

    @property
    def logs(self):
        task_logs = self.entity.task_logs
        if task_logs is not None:
            task_logs.seek(0)
            return task_logs.read()
        else:
            return super(CeleryTaskAdapter, self).logs

    @property
    def state(self):
        db_state = self.entity.cw_adapt_to('IWorkflowable').state
        db_final_state_map = {'done': STATES.SUCCESS, 'failed': STATES.FAILURE}
        if db_state in db_final_state_map:
            return db_final_state_map[db_state]
        elif task_in_backend(self.task_id):
            return super(CeleryTaskAdapter, self).state
        return _('unknown state')



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-celerytask automatic tests"""

import unittest

import celery.result

from cubicweb.devtools import testlib  # noqa

from cubes.celerytask.entities import (start_async_task, StartCeleryTaskOp,
                                       run_all_tasks)
from cubes.celerytask.testutils import BaseCeleryTaskTC

import cw_celerytask_helpers.redislogger as loghelper


class CeleryTaskTC(BaseCeleryTaskTC):

    def test_success_task(self):
        with self.admin_access.repo_cnx() as cnx:
            cwtask_eid = start_async_task(cnx, 'success', 42).eid
            cnx.commit()
            cwtask = cnx.entity_from_eid(cwtask_eid)
            run_all_tasks(cnx)
            self.wait_async_task(cnx, cwtask.task_id)
            result = cwtask.cw_adapt_to('ICeleryTask').result
            self.assertEqual(result.get(), 42)
            wf = cwtask.cw_adapt_to('IWorkflowable')
            self.assertEqual(wf.state, 'done')

    def test_fail_task(self):
        with self.admin_access.repo_cnx() as cnx:
            cwtask_eid = start_async_task(cnx, 'fail').eid
            cnx.commit()
            run_all_tasks(cnx)
            cwtask = cnx.entity_from_eid(cwtask_eid)
            self.wait_async_task(cnx, cwtask.task_id)
            result = cwtask.cw_adapt_to('ICeleryTask').result
            tb = result.traceback
            self.assertTrue(result.failed())
            self.assertTrue(tb.startswith('Traceback (most recent call last)'))
            self.assertTrue(tb.endswith('RuntimeError: fail\n'))
            wf = cwtask.cw_adapt_to('IWorkflowable')
            self.assertEqual(wf.state, 'failed')

    def test_api(self):
        with self.admin_access.repo_cnx() as cnx:
            run = cnx.create_entity('Run', name=u'success', i=12)
            cnx.commit()

            run = cnx.entity_from_eid(run.eid)
            self.assertIsNone(run.result)
            run.cw_adapt_to('IWorkflowable').fire_transition('start')
            cnx.commit()
            run_all_tasks(cnx)

            run = cnx.entity_from_eid(run.eid)
            self.wait_async_task(cnx, run.task_id)
            cnx.commit()
            self.assertEqual(run.cw_adapt_to('IWorkflowable').state, 'done')
            self.assertEqual(run.result, 12)

    def test_start_celerytask_op(self):
        with self.admin_access.repo_cnx() as cnx:
            # Task must run even if there is no data on current transaction
            task = celery.signature("success", kwargs={"n": 10})
            task_id = task.freeze().task_id
            cnx.transaction_data['celerytask'] = {42: task}
            StartCeleryTaskOp.get_instance(cnx).add_data(42)
            cnx.commit()
            run_all_tasks(cnx)
            result = self.wait_async_task(cnx, task_id)
            self.assertEqual(result.get(), 10)

    def test_logs(self):
        with self.admin_access.repo_cnx() as cnx:
            cwtask_eid = start_async_task(cnx, 'log').eid
            cnx.commit()
            run_all_tasks(cnx)
            cwtask = cnx.entity_from_eid(cwtask_eid)
            self.wait_async_task(cnx, cwtask.task_id)

            # logs should be flushed from redis to database
            redis_logs = loghelper.get_task_logs(cwtask.task_id)
            self.assertIsNone(redis_logs)
            cwtask.task_logs.seek(0)
            task_logs = cwtask.task_logs.read()
            logs = cwtask.cw_adapt_to('ICeleryTask').logs
            self.assertEqual(task_logs, logs)

            self.assertIn(b'out should be in logs', logs)
            self.assertIn(b'err should be in logs', logs)
            self.assertIn(b'cw warning should be in logs', logs)
            for name in (b'cw', b'celery'):
                for key in (b'error', b'critical', b'exception'):
                    self.assertIn(name + b' ' + key + b' should be in logs',
                                  logs)
            self.assertIn(b'cw critical should be in logs', logs)
            self.assertNotIn(b'should not be in logs', logs)
            self.assertIn(b'raise Exception("oops")', logs)

    @unittest.skipIf(celery.VERSION.major == 3, "not supported with celery 3")
    def test_workflow_chain(self):
        with self.admin_access.repo_cnx() as cnx:
            s = celery.signature
            task = celery.chain(s("add", (2, 2)), s("add", (4,)))
            cwtask = start_async_task(cnx, task)
            cnx.commit()
            run_all_tasks(cnx)
            result = self.wait_async_task(cnx, cwtask.task_id)
            self.assertEqual(result.get(), 8)

            children = cwtask.reverse_parent_task
            self.assertEqual(len(children), 1)
            result = children[0].cw_adapt_to('ICeleryTask').result
            self.assertEqual(result.get(), 4)

    def test_workflow_group(self):
        with self.admin_access.repo_cnx() as cnx:
            s = celery.signature
            task = celery.group(s("add", (2, 3)), s("add", (4, 5)))
            cwtask = start_async_task(cnx, task)
            cnx.commit()
            self.assertEqual(cwtask.task_name, u'celery.group')

            results = run_all_tasks(cnx)
            self.assertIn(cwtask.eid, results)
            self.assertEqual(results[cwtask.eid].get(), [5, 9])

            children = cwtask.reverse_parent_task
            self.assertEqual(len(children), 2)
            self.assertEqual(children[0].task_name, u'add')
            task_id = children[0].cw_adapt_to('ICeleryTask').task_id
            result = self.wait_async_task(cnx, task_id)
            self.assertEqual(result.get(), 5)

    def test_workflow_chord(self):
        with self.admin_access.repo_cnx() as cnx:
            s = celery.signature
            task = celery.chord([s("success", (i,)) for i in range(10)],
                                s("tsum", []))
            cwtask = start_async_task(cnx, task)
            cnx.commit()
            self.assertEqual(cwtask.task_name, u'celery.chord')
            run_all_tasks(cnx)
            result = self.wait_async_task(cnx, cwtask.task_id)
            self.assertEqual(result.get(), 45)

            children = cwtask.reverse_parent_task
            self.assertEqual([child.task_name for child in children],
                             [u'success'] * 10)
            self.assertCountEqual(
                [t.cw_adapt_to('ICeleryTask').result.get() for t in children],
                range(10))

    def test_workflow_subtasks(self):
        with self.admin_access.repo_cnx() as cnx:
            s = celery.signature
            task = s("spawn")
            cwtask = start_async_task(cnx, task)
            cnx.commit()
            self.assertEqual(cwtask.task_name, u'spawn')
            run_all_tasks()
            asresult = self.wait_async_task(cnx, cwtask.task_id)
            result = celery.result.from_serializable(asresult.result)
            self.assertEqual(0, result.get())

            if False:
                # require a running CeleryMonitor, and I don't know
                # how to get one running while executing the tests...
                children = cwtask.reverse_parent_task
                names = [child.task_name for child in children]
                self.assertEqual(10, len([x for x in names
                                          if x == 'success']))
                self.assertEqual(1, len([x for x in names
                                         if x == 'tsum']))
                self.assertEqual(1, len([x for x in names
                                         if x == 'add']))


if __name__ == '__main__':
    from unittest import main
    main()




from cubes.celerytask.entities import start_async_task

tasks = []
# a very simple task
tasks.append(start_async_task(cnx, 'hi_there', 'THERE', kw=42))

# a task showing several aspects of logging in cubicweb-celerytask
tasks.append(start_async_task(cnx, 'log'))

# a task with a progress bar
tasks.append(start_async_task(cnx, 'show-me-progress'))

cnx.commit()

print ("Started tasks:")
print ('\n'.join('  %s'%t.cwuri for t in tasks))



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-celerytask specific hooks and operations"""
from __future__ import print_function

import six
from celery import current_app

from cubicweb import Binary
from cubicweb.predicates import on_fire_transition
from cubicweb.server.hook import Hook, DataOperationMixIn, Operation

from cw_celerytask_helpers.redislogger import flush_task_logs

if six.PY3:
    unicode = six.text_type


class FlushCeleryTaskLogsOp(DataOperationMixIn, Operation):

    def postcommit_event(self):
        for task_id in self.get_data():
            flush_task_logs(task_id)


class CeleryTaskFinishedHook(Hook):
    __regid__ = 'celerytask.celerytask_finished'
    __select__ = (Hook.__select__ &
                  on_fire_transition('CeleryTask', ('finish', 'fail')))
    events = ('after_add_entity',)

    def __call__(self):
        if current_app.conf.CELERY_ALWAYS_EAGER:
            return
        entity = self.entity.for_entity
        logs = Binary(entity.cw_adapt_to('ICeleryTask').logs)
        entity.cw_set(task_logs=logs)
        FlushCeleryTaskLogsOp.get_instance(self._cw).add_data(entity.task_id)



from __future__ import print_function

import sys
import logging
import time
import random

from celery import current_app as app
from celery import chord
from celery.utils.log import get_task_logger
from cw_celerytask_helpers.redislogger import redirect_stdouts

# 3 kinds of logger
cw_logger = logging.getLogger('cubes.tasks')
dummy_logger = logging.getLogger('dummy')
logger = get_task_logger(__name__)


@app.task(name='success')
def success(n):
    logger.info('TASK SUCCESS: (%s)', n)
    time.sleep(random.random()*2)
    return n


@app.task(name='fail')
def fail():
    logger.info('TASK FAIL')
    raise RuntimeError('fail')


@app.task(name='log')
def log():
    for out in [sys.stdout, sys.stderr]:
        print('should not be in logs', file=out)

    with redirect_stdouts(logger):
        print('out should be in logs')
        print('err should be in logs', file=sys.stderr)

    for out in [sys.stdout, sys.stderr]:
        print('should not be in logs', file=out)

    for name, l, state in [
        ('cw', cw_logger, 'be'),
        ('celery', logger, 'be'),
        ('dummy', dummy_logger, 'not be')
    ]:
        for key in ('debug', 'info', 'warning', 'error', 'critical'):
            getattr(l, key)('%s %s should %s in logs', name, key, state)
        try:
            raise RuntimeError("fail")
        except RuntimeError:
            l.exception('%s exception should be in logs', name)

    raise Exception("oops")


@app.task(name="add")
def add(x, y):
    logger.info('TASK ADD: (%s, %s)', x, y)
    return x + y


@app.task(bind=True, name="tsum")
def tsum(self, args, root_id=None):
    if root_id is None:
        root_id = self.request.id
    logger.info('TASK SUM: (%s) <root=%s>', len(args), root_id)
    return sum(args)


@app.task(bind=True, name='hi_there')
def my_task(self, arg, kw=0, root_id=None):
    if root_id is None:
        root_id = self.request.id
    logger.info('HI %s %s! <root=%s>', arg, kw, root_id)
    if kw is 'wait':
        time.sleep(1)
    return 42


@app.task(bind=True, name='show-me-progress')
def progress(self, steps=10):
    self.update_state(state='PROGRESS', meta={'progress': 0.0})
    for i in range(steps):
        val = random.random()*2
        logger.info('Step %s: going to sleep for %.2fs', i, val)
        time.sleep(val)
        self.update_state(state='PROGRESS',
                          meta={'progress': (i+1/steps)})


@app.task(bind=True, name='multi')
def multi(self):
    logger.info('Starting multi!')
    time.sleep(2)
    logger.info('Running subtasks!')
    r = chord([my_task.si(i, kw='wait') for i in range(5)],
              tsum.s())
    logger.info('multi done!')
    return r().serializable()



# pylint: disable=W0622
"""cubicweb-celerytask application packaging information"""

from os import listdir as _listdir
from os.path import join, isdir
from glob import glob


modname = 'celerytask'
distname = 'cubicweb-celerytask'

numversion = (0, 3, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = 'Run and monitor celery tasks'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'cubicweb': '>= 3.19',
    'six': '>= 1.4.0',
    'celery': '< 4.0.0',
    'cw-celerytask-helpers': '>= 0.2.0',
}
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
]

THIS_CUBE_DIR = join('share', 'cubicweb', 'cubes', modname)


def listdir(dirpath):
    return [join(dirpath, fname) for fname in _listdir(dirpath)
            if fname[0] != '.' and not fname.endswith('.pyc') and
            not fname.endswith('~') and
            not isdir(join(dirpath, fname))]


data_files = [
    # common files
    [THIS_CUBE_DIR, [fname for fname in glob('*.py') if fname != 'setup.py']],
]
# check for possible extended cube layout
for dname in ('entities', 'views', 'sobjects', 'hooks', 'schema', 'data',
              'wdoc', 'i18n', 'migration'):
    if isdir(dname):
        data_files.append([join(THIS_CUBE_DIR, dname), listdir(dname)])
# Note: here, you'll need to add subdirectories if you want
# them to be included in the debian package
